﻿' See "Enhanced Fireworks Algorithm", S. Zheng, A. Janecek, Y. Tan (2013)
Namespace FireworksAlgorithm
	Friend Class FireworksProgram
		Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbCrLf & "Begin fireworks algorithm optimization demo" & vbCrLf)
			Console.WriteLine("Goal is to find solution to Ackley's function for 10 variables")
			Console.WriteLine("Function has known min value = 0.0 at (0, 0, 0, 0, 0, 0, 0, 0, 0, 0)")

			Dim [dim] As Integer = 10 ' often number weights to find in ML training
			Dim n As Integer = 5 ' number fireworks
			Dim maxEpochs As Integer = 1000
            Console.WriteLine(vbCrLf & "Setting Ackley's dimension to " & [dim])
			Console.WriteLine("Setting number fireworks to " & n)
			Console.WriteLine("Setting maxEpochs to " & maxEpochs)

            Console.WriteLine(vbCrLf & "Begin algorithm" & vbCrLf)
			Dim bestPosition() As Double = Solve([dim], n, maxEpochs) ' use fireworks algorithm
            Console.WriteLine(vbCrLf & "Algorithm complete")

            Console.WriteLine(vbCrLf & "Best solution found: ")
            For i = 0 To [dim] - 1
                Console.Write(bestPosition(i).ToString("F3") & " ")
            Next i
			Console.WriteLine()

			Dim [error] As Double = FireworksProgram.Error(bestPosition)
            Console.WriteLine(vbCrLf & "Error of best solution found = " & [error].ToString("F5"))

            Console.WriteLine(vbCrLf & "End fireworks algorithm optimization demo" & vbCrLf)
			Console.ReadLine()
		End Sub ' Main

		Public Shared Function [Error](ByVal position() As Double) As Double
			' Ackley's function has min = 0.0 at (0.0, 0.0, . . 0.0)
			Dim ss As Double = 0.0
            For i = 0 To position.Length - 1
                ss = ss + (position(i) * position(i))
            Next i
			Dim mss As Double = ss / position.Length

			Dim sc As Double = 0.0
            For i = 0 To position.Length - 1
                sc = sc + Math.Cos(2.0 * Math.PI * position(i))
            Next i
			Dim msc As Double = sc / position.Length

			Dim f As Double = -20.0 * Math.Exp(-0.2 * Math.Sqrt(mss)) - Math.Exp(msc) + 20.0 + Math.E
			Dim trueMinVal As Double = 0.0
			'return Math.Abs(f - trueMinVal);
			Return (f - trueMinVal) * (f - trueMinVal)
		End Function

		Public Shared Function Solve(ByVal [dim] As Integer, ByVal n As Integer, ByVal maxEpochs As Integer) As Double()
			Dim m As Integer = n * 10 ' total number of regular sparks for all fireworks
			Dim mHat As Integer = 5 ' number of Gaussian sparks
			Dim a As Double = 0.04 ' controls min sparks per firework
			Dim b As Double = 0.8 ' controls max sparks per firework
'INSTANT VB NOTE: The variable A was renamed since Visual Basic will not allow local variables with the same name as parameters or other local variables:
			Dim A_Renamed As Integer = 40 ' max amplitude
			Dim minX As Double = -10.0
			Dim maxX As Double = 10.0

			' generate n random fireworks
			Dim rnd As New Random(3) ' seed = 3 gives representative demo
			Dim fireworks(n - 1) As Info
            For i = 0 To n - 1
                fireworks(i) = New Info()
                fireworks(i).position = New Double([dim] - 1) {}
                For j = 0 To [dim] - 1
                    fireworks(i).position(j) = (maxX - minX) * rnd.NextDouble() + minX
                Next j
                fireworks(i).error = [Error](fireworks(i).position)
            Next i

			' initialize best position, error to dummy values
			Dim bestPosition([dim] - 1) As Double
            For k = 0 To [dim] - 1
                bestPosition(k) = fireworks(0).position(k)
            Next k
			Dim bestError As Double = fireworks(0).error ' arbitrary

			Dim sparksList(n - 1) As List(Of Info) ' each cell is a ref to a List of Info objects.
            For i = 0 To n - 1
                sparksList(i) = New List(Of Info)() ' empty List
            Next i

			' main processing loop
			Dim epoch As Integer = 0
			Do While epoch < maxEpochs
				If epoch Mod 100 = 0 Then ' show progress every 100 iterations
					Console.Write("epoch = " & epoch)
					Console.WriteLine(" error at best known position = " & bestError.ToString("F4"))
				End If

				Dim numberSparks() As Integer = FireworksProgram.NumberSparks(fireworks, m, a, b) ' number regular sparks for each firework
				Dim amplitudes() As Double = FireworksProgram.Amplitudes(fireworks, A_Renamed, epoch, maxEpochs, minX, maxX) ' amplitude each firework. epoch, maxEpochs, minX, maxX to establish a min amplitude

                For i = 0 To n - 1
                    sparksList(i).Clear() ' number of sparks changes each epoch
                Next i

                For i = 0 To n - 1 ' generate regular sparks for each firework
                    Dim amp As Double = amplitudes(i) ' amplitude for curr firework
                    Dim ns As Integer = numberSparks(i) ' number sparks for curr firework

                    For j = 0 To ns - 1 ' each spark for curr firework
                        Dim spark As New Info() ' a spark has a position and error
                        spark.position = New Double([dim] - 1) {} ' allocate space (ctor doesn't)
                        For k = 0 To [dim] - 1 ' spark position based on its parent firework
                            spark.position(k) = fireworks(i).position(k)
                        Next k
                        Dim z As Integer = CInt(Fix(Math.Round([dim] * rnd.NextDouble))) ' number of random dimensions
                        Dim dimensions() As Integer = PickDimensions([dim], z, epoch) ' select z random dimensions. epoch is a rnd seed
                        For ii = 0 To dimensions.Length - 1 ' each of the randomly selected dimensions
                            Dim h As Double = amp * 2 * rnd.NextDouble - 1 ' displacement hi = +1, lo = -1, (hi - lo) * r + lo
                            Dim k As Integer = dimensions(ii) ' convenience
                            spark.position(k) += h ' displace from parent firework
                            If spark.position(k) < minX OrElse spark.position(k) > maxX Then ' bring out-of-range values back in
                                spark.position(k) = (maxX - minX) * rnd.NextDouble + minX
                            End If
                        Next ii
                        spark.error = [Error](spark.position)
                        sparksList(i).Add(spark)

                        ' is curr spark global best?
                        If spark.error < bestError Then
                            bestError = spark.error
                            For k = 0 To [dim] - 1
                                bestPosition(k) = spark.position(k)
                            Next k
                        End If
                    Next j ' each new regular spark
                Next i ' each firework

				' pretty hideous parameter passing . . 
				AddGaussianSparks(fireworks, sparksList, [dim], mHat, epoch, minX, maxX, bestPosition, bestError, rnd)

				' now pick numFireworks new fireworks from all sparks
				' use best spark, worst spark, and n-2 random sparks
				' find best and worst spark
				Dim bestSparkPos([dim] - 1) As Double
				Dim bestSparkErr As Double = Double.MaxValue

				Dim worstSparkPos([dim] - 1) As Double
				Dim worstSparkErr As Double = Double.MinValue

                For i = 0 To n - 1 ' numFireworks
                    Dim j As Integer = 0
                    Do While j < sparksList(i).Count ' number sparks in each firework
                        If sparksList(i)(j).error < bestSparkErr Then
                            bestSparkErr = sparksList(i)(j).error
                            For k = 0 To sparksList(i)(j).position.Length - 1
                                bestSparkPos(k) = sparksList(i)(j).position(k)
                            Next k
                        End If
                        If sparksList(i)(j).error > worstSparkErr Then
                            worstSparkErr = sparksList(i)(j).error
                            For k = 0 To sparksList(i)(j).position.Length - 1
                                worstSparkPos(k) = sparksList(i)(j).position(k)
                            Next k
                        End If
                        j += 1
                    Loop ' each spark
                Next i ' each firework

                For k = 0 To [dim] - 1 ' first new firework is best spark
                    fireworks(0).position(k) = bestSparkPos(k)
                Next k
				fireworks(0).error = bestSparkErr

                For k = 0 To [dim] - 1 ' second new firework is worst spark
                    fireworks(1).position(k) = worstSparkPos(k)
                Next k
				fireworks(1).error = worstSparkErr

                For i = 2 To n - 1 ' n-2 random sparks
                    Dim row As Integer = rnd.Next(0, n) ' more likely to be a good spark
                    Dim cols As Integer = sparksList(row).Count
                    Dim col As Integer = rnd.Next(0, cols)
                    For k = 0 To [dim] - 1
                        fireworks(i).position(k) = sparksList(row)(col).position(k)
                    Next k
                    fireworks(i).error = sparksList(row)(col).error
                Next i
				' consider, instead selecting a row based on row-count . . 

				epoch += 1
			Loop ' main loop
			Return bestPosition
		End Function ' Solve

		'private static void ShowVector(int[] vector)
		'{
		'  Console.WriteLine("");
		'  for (int i = 0; i < vector.Length; ++i)
		'    Console.Write(vector[i] + " ");
		'  Console.WriteLine("\n");
		'}

		'private static void ShowInfo(Info info)
		'{
		'  Console.WriteLine("iiiiiiiiiiiiiiiiiiiiiiiiiiiiii");
		'  Console.Write(" pos = ");
		'  for (int k = 0; k < info.position.Length; ++k)
		'    Console.Write(info.position[k].ToString("F2") + " ");
		'  Console.WriteLine("  error = " + info.error.ToString("F3"));

		'  Console.WriteLine("iiiiiiiiiiiiiiiiiiiiiiiiiiiiii");
		'}

		'private static void ShowFireworks(Info[] fireworks)
		'{
		'  Console.WriteLine("ffffffffffffffffffffffffffffff");
		'  for (int i = 0; i < fireworks.Length; ++i)
		'  {
		'    Console.Write(i + " pos = ");
		'    for (int k = 0; k < fireworks[i].position.Length; ++k)
		'      Console.Write(fireworks[i].position[k].ToString("F2") + " ");
		'    Console.WriteLine("  error = " + fireworks[i].error.ToString("F3"));
		'  }
		'  Console.WriteLine("ffffffffffffffffffffffffffffff");
		'}

		'private static void ShowSparksList(List<Info>[] sparks)
		'{
		'  Console.WriteLine("++++++++++++++++++++++++++++");
		'  for (int i = 0; i < sparks.Length; ++i)
		'  {
		'    Console.Write(i + " ");
		'    for (int j = 0; j < sparks[i].Count; ++j)
		'    {
		'      Console.Write(" { ");
		'      for (int k = 0; k < sparks[i][j].position.Length; ++k)
		'        Console.Write(sparks[i][j].position[k].ToString("F2") + " ");
		'      Console.Write(" } ");
		'      Console.WriteLine(" err = " + sparks[i][j].error.ToString("F3"));
		'    }
		'  }
		'  Console.WriteLine("++++++++++++++++++++++++++++");
		'}

		Private Shared Function PickDimensions(ByVal [dim] As Integer, ByVal z As Integer, ByVal seed As Integer) As Integer()
			' pick z random dimensions of a position
			Dim result(z - 1) As Integer
			Dim indices([dim] - 1) As Integer
            For i = 0 To [dim] - 1
                indices(i) = i
            Next i

			Dim rnd As New Random(seed) ' shuffle indices
            For i = 0 To indices.Length - 1
                Dim ri As Integer = rnd.Next(i, indices.Length)
                Dim tmp As Integer = indices(ri)
                indices(ri) = indices(i)
                indices(i) = tmp
            Next i
			' copy first z indices to result
            For i = 0 To z - 1
                result(i) = indices(i)
            Next i

			Return result
		End Function

		Private Shared Function YMax(ByVal fireworks() As Info) As Double
			' largest (worst) error in any firework
			Dim result As Double = fireworks(0).error
            For i = 1 To fireworks.Length - 1
                If fireworks(i).error > result Then
                    result = fireworks(i).error
                End If
            Next i
			Return result
		End Function

		Private Shared Function YMin(ByVal fireworks() As Info) As Double
			' smallest (best) error in any firework
			Dim result As Double = fireworks(0).error
            For i = 1 To fireworks.Length - 1
                If fireworks(i).error < result Then
                    result = fireworks(i).error
                End If
            Next i
			Return result
		End Function

		Private Shared Function NumberSparks(ByVal fireworks() As Info, ByVal m As Integer, ByVal a As Double, ByVal b As Double) As Integer()
			' number sparks for each firework
			Dim n As Integer = fireworks.Length
			Dim minSparks As Integer = CInt(Fix(Math.Round(a * m))) ' if n=5, m=50, a=.04, -> 2
			If minSparks < 1 Then
				minSparks = 1
			End If
			Dim maxSparks As Integer = CInt(Fix(Math.Round(b * m))) ' if n=5, m=50, b=.8 -> 40
			If maxSparks > m - (n - 1) * minSparks Then
				maxSparks = m - (n - 1) * minSparks
			End If

			Dim yMax As Double = FireworksProgram.YMax(fireworks)
			Dim sumDeltas As Double = 0.0 ' sum diffs between yMax and each error
            For i = 0 To n - 1
                sumDeltas += yMax - fireworks(i).error
            Next i

			Dim numSparks(n - 1) As Integer ' the result
            For i = 0 To n - 1
                numSparks(i) = CInt(Fix(Math.Round(m * (yMax - fireworks(i).error + 0.0000000001) / (sumDeltas + 0.0000000001))))
                If numSparks(i) < minSparks Then
                    numSparks(i) = minSparks
                ElseIf numSparks(i) > maxSparks Then
                    numSparks(i) = maxSparks
                End If
            Next i
			Return numSparks
		End Function

		Private Shared Function Amplitudes(ByVal fireworks() As Info, ByVal A As Integer, ByVal epoch As Integer, ByVal maxEpochs As Integer, ByVal minX As Double, ByVal maxX As Double) As Double()
			Dim n As Integer = fireworks.Length
			Dim yMin As Double = FireworksProgram.YMin(fireworks)
			Dim sumDeltas As Double = 0.0 ' sum diffs between yMin and each error
            For i = 0 To n - 1
                sumDeltas += fireworks(i).error - yMin
            Next i

			Dim result(n - 1) As Double ' an amplitude for each firework
			Dim minAmplitude As Double = FireworksProgram.MinAmplitude(epoch, maxEpochs, minX, maxX)
            For i = 0 To n - 1
                result(i) = A * (fireworks(i).error - yMin + 0.0000000001) / (sumDeltas + 0.0000000001)
                If result(i) < minAmplitude Then
                    result(i) = minAmplitude
                End If
            Next i
			Return result
		End Function

		Private Shared Function MinAmplitude(ByVal epoch As Integer, ByVal maxEpochs As Integer, ByVal minX As Double, ByVal maxX As Double) As Double
			' minimum amplitude for any firework at curr epoch 
			Dim Ainit As Double = (maxX - minX) * 0.02
			Dim Afinal As Double = (maxX - minX) * 0.001
			Return Ainit - (Ainit - Afinal) * epoch \ maxEpochs
		End Function

		Private Shared Sub AddGaussianSparks(ByVal fireworks() As Info, ByVal sparksList() As List(Of Info), ByVal [dim] As Integer, ByVal mHat As Integer, ByVal epoch As Integer, ByVal minX As Double, ByVal maxX As Double, ByVal bestPosition() As Double, ByRef bestError As Double, ByVal rnd As Random)
			' generate mHat Gaussian sparks, add to sparksList, update bestPOsition[], bestError
			Dim n As Integer = fireworks.Length
            For g = 0 To mHat - 1
                Dim gSpark As New Info
                gSpark.position = New Double([dim] - 1) {}

                Dim i As Integer = rnd.Next(0, n) ' pick a random firework
                For k = 0 To [dim] - 1 ' spark position based on its parent firework
                    gSpark.position(k) = fireworks(i).position(k)
                Next k
                Dim z As Integer = CInt(Fix(Math.Round([dim] * rnd.NextDouble))) ' number of random dimensions
                Dim dimensions() As Integer = PickDimensions([dim], z, epoch) ' pick random dimensions
                Dim u1 As Double = rnd.NextDouble ' make Gaussian displacement u = 0, sd = 1
                Dim u2 As Double = rnd.NextDouble
                Dim left As Double = Math.Cos(2.0 * Math.PI * u1)
                Dim right As Double = Math.Sqrt(-2.0 * Math.Log(u2))
                Dim e As Double = left * right ' mean = 0, sd = 1
                For ii = 0 To dimensions.Length - 1 ' each of the randomly selected dimensions
                    Dim k As Integer = dimensions(ii) ' convenience
                    gSpark.position(k) = gSpark.position(k) + (bestPosition(k) - gSpark.position(k)) * e
                    If gSpark.position(k) < minX OrElse gSpark.position(k) > maxX Then ' bring out-of-range values back in
                        gSpark.position(k) = (maxX - minX) * rnd.NextDouble() + minX
                    End If
                Next ii
                gSpark.error = [Error](gSpark.position)
                sparksList(i).Add(gSpark)

                If gSpark.error < bestError Then
                    bestError = gSpark.error
                    For k = 0 To [dim] - 1
                        bestPosition(k) = gSpark.position(k)
                    Next k
                End If
            Next g ' each Gaussian spark
		End Sub

	End Class ' Program

	Public Class Info
		Public position() As Double
		Public [error] As Double
	End Class

End Namespace ' ns
